import logging
from logging import getLogger, INFO, DEBUG


def get_logger(debug=False):
    log_fmt = "%(asctime)s %(levelname)s - %(message)s"
    logging.basicConfig(format=log_fmt)

    logger = getLogger(__name__)
    if debug:
        logger.setLevel(DEBUG)
    else:
        logger.setLevel(INFO)
    return logger



import click
import json
from datetime import datetime
from .billing import Billing
from .cost_explorer import CostExplorer
from .util import Util
from .validator import Validator
from .date_util import DateUtil
from . import constants


class Mash(object):
    pass


@click.group(invoke_without_command=True)
@click.option(
    "--debug/--no-debug", default=False, help="enable debug logging. (default: False)"
)
@click.option("--profile", type=str, help="aws profile name.")
@click.pass_context
def cli(ctx, debug, profile):
    ctx.obj = Mash()
    ctx.obj.debug = debug
    ctx.obj.profile = profile


@cli.command(help="list cost explorer")
@click.option(
    "--granularity",
    "-g",
    type=click.Choice(["DAILY", "MONTHLY"]),
    default="MONTHLY",
    help="granularity. (default: MONTHLY)",
)
@click.option(
    "--point",
    "-p",
    type=int,
    default=constants.DEFAULT_POINT,
    help=f"duration. if granularity is MONTHLY, {constants.DEFAULT_POINT} month ago start. if granularity is DAILY, {constants.DEFAULT_POINT} day ago start. (default: {constants.DEFAULT_POINT})",
)
@click.option(
    "--start",
    callback=Validator.validate_dateformat,
    type=str,
    help=f"range of start day. default is {constants.DEFAULT_POINT} month ago.",
)
@click.option(
    "--end",
    callback=Validator.validate_dateformat,
    type=str,
    default=datetime.today().strftime("%Y-%m-%d"),
    help="range of end day. default is now.",
)
@click.option(
    "--tablefmt",
    "-t",
    type=str,
    default="simple",
    help="tabulate format. (default: simple)",
)
@click.option(
    "--group-by",
    type=click.Choice(constants.AVAILABLE_DIMENSIONS),
    multiple=True,
    default=["SERVICE"],
    help='group by keys. (default: ["SERVICE"])',
)
@click.option(
    "--filter", type=json.loads, help="filter of dimensions. default is no filter."
)
@click.option(
    "--metrics",
    type=click.Choice(constants.AVAILABLE_METRICS),
    default=constants.DEFAULT_METRICS,
    help="metrics. (default: UnblendedCost)",
)
@click.pass_context
def list_ce(ctx, granularity, point, start, end, tablefmt, group_by, filter, metrics):
    cost_explorer = CostExplorer(
        granularity,
        start or DateUtil.get_start(granularity, point),
        end,
        group_by=group_by,
        filter_dimensions=filter,
        metrics=metrics,
        debug=ctx.obj.debug,
        profile=ctx.obj.profile,
    )
    currencies = cost_explorer.get_cost_and_usage_total_and_group_by()
    print(Util.convert_tabulate(currencies, tablefmt=tablefmt))


@cli.command(help="list cloudwatch billing")
@click.option(
    "--range", "-r", type=click.Choice(["month", "week", "day"]), required=True
)
@click.option(
    "--tablefmt",
    "-t",
    type=str,
    default="simple",
    help="tabulate format. (default: simple)",
)
@click.option(
    "--point",
    "-p",
    type=int,
    default=constants.DEFAULT_POINT,
    help=f"number of data point. (default: {constants.DEFAULT_POINT})",
)
@click.pass_context
def list_billing(ctx, range, tablefmt, point):
    billing = Billing(ctx.obj.debug, ctx.obj.profile)
    currencies = billing.get_currencies_per_service(range, point)
    print(Util.convert_tabulate(currencies, tablefmt=tablefmt))


def main():
    cli(obj={})



from tabulate import tabulate


class Util:
    @staticmethod
    def convert_tabulate(results, tablefmt="simple"):
        converts = []
        for k, amounts in results.items():
            converts.append(dict({"key": k}, **amounts))
        last_time = list(converts[0].keys())[-1]
        converts = sorted(
            converts,
            key=lambda x: 0 if x.get(last_time) is None else x.get(last_time),
            reverse=True,
        )
        return tabulate(converts, headers="keys", tablefmt=tablefmt)



import click
from datetime import datetime


class Validator:
    @staticmethod
    def validate_dateformat(ctx, param, value):
        if value is None:
            return value

        try:
            datetime.strptime(value, "%Y-%m-%d")
        except:
            raise click.BadParameter(f"please input dateformat parameter. ('%Y-%m-%d')")
        return value



from boto3.session import Session
from .logger import get_logger


class CostExplorerClient:
    def __init__(
        self,
        granularity,
        start,
        end,
        filter_dimensions=None,
        metrics=None,
        profile=None,
        debug=False,
    ):
        self.granularity = granularity
        self.start = start
        self.end = end
        self.filter_dimensions = filter_dimensions
        self.metrics = metrics
        self.client = Session(profile_name=profile).client(
            "ce", region_name="us-east-1"
        )
        self.logger = get_logger(debug=debug)

    def get_cost_and_usage(self, group_by=None):
        """
        cost explorerのAPIを実行
        """
        params = dict(
            TimePeriod={"Start": self.start, "End": self.end},
            Granularity=self.granularity,
            Metrics=[self.metrics,],
        )
        group_by = self._get_group_by(group_by)
        if group_by is not None:
            params["GroupBy"] = group_by
        if self.filter_dimensions is not None:
            params["Filter"] = {"Dimensions": self.filter_dimensions}
        self.logger.debug(params)
        response = self.client.get_cost_and_usage(**params)
        return response.get("ResultsByTime")

    def _get_group_by(self, group_by):
        """
        listからgroup-byの構造に変換
        """
        if group_by is None:
            return None
        return [{"Type": "DIMENSION", "Key": key} for key in group_by]



from datetime import datetime
from .logger import get_logger
from .cost_explorer_client import CostExplorerClient
from . import constants


class CostExplorer:
    def __init__(
        self,
        granularity,
        start,
        end,
        group_by=[],
        filter_dimensions=None,
        metrics=constants.DEFAULT_METRICS,
        profile=None,
        debug=False,
    ):
        self.cost_explorer_client = CostExplorerClient(
            granularity,
            start,
            end,
            filter_dimensions=filter_dimensions,
            metrics=metrics,
            profile=profile,
            debug=debug,
        )
        self.granularity = granularity
        self.group_by = group_by
        self.metrics = metrics
        self.logger = get_logger(debug=debug)

    def get_cost_and_usage_total_and_group_by(self):
        """
        startとendを指定してcostの時系列データを取得し、totalとgroup_byをmergeする
        """
        # totalを取得
        total = self.get_cost_and_usage_total()

        # group別を取得
        results = self.get_cost_and_usage_group_by()

        # totalとgroup byをmergeする
        merged = dict(total, **results)
        self.logger.debug(merged)
        return merged

    def get_cost_and_usage_total(self):
        """
        startとendを指定してcostの時系列データを取得する
        """
        # totalを取得
        cost_and_usage = self.cost_explorer_client.get_cost_and_usage()
        total = self._convert_results_total(cost_and_usage)
        self.logger.debug(total)
        return total

    def get_cost_and_usage_group_by(self):
        """
        startとendを指定してcostの時系列データを取得する
        """
        cost_and_usage_per_service = self.cost_explorer_client.get_cost_and_usage(
            group_by=self.group_by
        )
        results = self._convert_results_group_by(cost_and_usage_per_service)
        self.logger.debug(results)
        return results

    def _convert_results_group_by(self, cost_and_usage_per_service):
        """
        group-byが指定されているデータ構造のparse
        """
        results = {}
        for result in cost_and_usage_per_service:
            start_period = result.get("TimePeriod").get("Start")
            time_key = self._convert_period(start_period)
            groups = result.get("Groups")
            for group in groups:
                group_by_key = ",".join(group.get("Keys"))
                if results.get(group_by_key) is None:
                    results[group_by_key] = {}
                else:
                    results[group_by_key] = results.get(group_by_key)
                metrics = group.get("Metrics")
                amount = metrics.get(self.metrics).get("Amount")
                results[group_by_key][time_key] = round(float(amount), 2)
        return results

    def _convert_results_total(self, cost_and_usage_per_service):
        """
        Totalのデータ構造のparse
        """
        results = {"Total": {}}
        for result in cost_and_usage_per_service:
            start_period = result.get("TimePeriod").get("Start")
            time_key = self._convert_period(start_period)
            metrics = result.get("Total")
            amount = metrics.get(self.metrics).get("Amount")
            results["Total"][time_key] = round(float(amount), 2)
        return results

    def _convert_period(self, start_period):
        """
        headerに表示させる日付をmonthlyとdailyで切り替え
        """
        if self.granularity == "MONTHLY":
            return datetime.strptime(start_period, "%Y-%m-%d").strftime("%Y-%m")
        return datetime.strptime(start_period, "%Y-%m-%d").strftime("%m-%d")



from datetime import timedelta
from datetime import datetime


class DateUtil:
    @staticmethod
    def get_start(granularity, point):
        """
        datapointのscaleをmonth, week, dayで自動調節する
        """
        if granularity == "MONTHLY":
            days = 30 * point
        elif granularity == "DAILY":
            days = point

        start_datetime = datetime.today() - timedelta(days=days)
        if granularity == "MONTHLY":
            start_datetime = start_datetime.replace(day=1)
        return start_datetime.strftime("%Y-%m-%d")



AVAILABLE_DIMENSIONS = [
    "AZ",
    "INSTANCE_TYPE",
    "LINKED_ACCOUNT",
    "OPERATION",
    "PURCHASE_TYPE",
    "SERVICE",
    "USAGE_TYPE",
    "PLATFORM",
    "TENANCY",
    "RECORD_TYPE",
    "LEGAL_ENTITY_NAME",
    "DEPLOYMENT_OPTION",
    "DATABASE_ENGINE",
    "CACHE_ENGINE",
    "INSTANCE_TYPE_FAMILY",
    "REGION",
    "BILLING_ENTITY",
    "RESERVATION_ID",
    "SAVINGS_PLANS_TYPE",
    "SAVINGS_PLAN_ARN",
    "OPERATING_SYSTEM",
]

AVAILABLE_METRICS = [
    "BlendedCost",
    "UnblendedCost",
    "AmortizedCost",
    "NetAmortizedCost",
    "NetUnblendedCost",
    "UsageQuantity",
    "NormalizedUsageAmount",
]

DEFAULT_METRICS = "UnblendedCost"
DEFAULT_POINT = 7



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))
long_description = open(os.path.join(here, "README.md")).read()

version = "0.0.4"

install_requires = ["tabulate", "boto3", "click>=7.0"]

setup(
    name="awscost",
    scripts=["bin/awscost"],
    version=version,
    description="Command Line utility for Amazon Aurora.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="awscost tool",
    author="Hiroshi Toyama",
    author_email="toyama0919@gmail.com",
    url="https://github.com/toyama0919/awscost",
    license="MIT",
    packages=find_packages("src", exclude=["tests"]),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={"console_scripts": ["awscost=awscost.commands:main"]},
)



from .logger import get_logger
from datetime import datetime
from dateutil.relativedelta import relativedelta
from boto3.session import Session


class Billing:
    def __init__(self, debug, profile):
        self.client = Session(profile_name=profile).client(
            "cloudwatch", region_name="us-east-1"
        )
        self.logger = get_logger(debug)
        response = self.client.list_metrics(Namespace="AWS/Billing")
        self.metrics = response.get("Metrics")

    def get_currencies_per_service(self, scale_range, point):
        """
        直近のサービス別のcostの時系列データを取得する
        """
        dateformat, period, delta = self._get_scale(scale_range, point)
        start_time = datetime.today() - relativedelta(days=delta)
        end_time = datetime.today()
        result = {}
        for metric in self.metrics:
            dimensions = metric.get("Dimensions")

            datapoints = self._get_datapoints(dimensions, period, start_time, end_time)
            service_name = self._get_service_name(dimensions)
            self.logger.debug(service_name)
            row = self._get_service_cost(datapoints, metric, dateformat)
            result[service_name] = row
        return result

    def _get_datapoints(self, dimensions, period, start_time, end_time):
        """
        datapoints(サービス単位のcostの時系列データ)を取得する
        """
        response = self.client.get_metric_statistics(
            Namespace="AWS/Billing",
            MetricName="EstimatedCharges",
            Dimensions=dimensions,
            StartTime=start_time,
            EndTime=end_time,
            Period=period,
            Statistics=["Maximum"],
        )
        return sorted(response.get("Datapoints"), key=lambda x: x.get("Timestamp"))

    def _get_service_name(self, dimensions):
        """
        service名をdimensionsから取得する。
        service名がdimensionsに入っていない場合はTotalになる
        """
        services = [
            dimension
            for dimension in dimensions
            if dimension.get("Name") == "ServiceName"
        ]
        return services[0].get("Value") if len(services) != 0 else "Total"

    def _get_service_cost(self, datapoints, metric, dateformat):
        """
        サービスごとのcostを取得する
        """
        row = {}
        for datapoint in datapoints:
            timestamp = datapoint.get("Timestamp").strftime(dateformat)
            maximum = datapoint.get("Maximum")
            row[timestamp] = maximum
        return row

    def _get_scale(self, scale_range, point):
        """
        datapointのscaleをmonth, week, dayで自動調節する
        """
        if scale_range == "month":
            return "%Y-%m", 60 * 60 * 24 * 30, 30 * point
        elif scale_range == "week":
            return "%m-%d", 60 * 60 * 24 * 7, 7 * point
        elif scale_range == "day":
            return "%m-%d", 60 * 60 * 24, 1 * point
        return "%m/%d", 86400, 1 * point
